"""Config Dictionary Dumpers."""

import json

from .__optional_imports import try_import
from .__typing import ConfigDict, FilePath


def json_dumper(config: ConfigDict, location: FilePath) -> None:
    """Dump config to JSON file.

    Args:
        config (ConfigDict): configuration
        location (FilePath): file to write
    """
    with open(location, "w", encoding="utf-8") as _file:
        json.dump(config, _file)


def yaml_dumper(config: ConfigDict, location: FilePath) -> None:
    """Dump config to YAML file.

    Args:
        config (ConfigDict): configuration
        location (FilePath): file to write

    Raises:
        ModuleNotFoundError: pyyaml is required
    """

    yaml = try_import("yaml")

    if yaml is None:  # pragma: no cover
        message = "Please install the pyyaml library."
        raise ModuleNotFoundError(message)

    with open(location, "w", encoding="utf-8") as _file:
        # NOTE: we must convert config from OrderedDict to dict because
        # pyyaml can't load OrderedDict for python <= 3.8
        yaml.dump(dict(config), _file)


def toml_dumper(config: ConfigDict, location: FilePath) -> None:
    """Dump config to TOML file.

    Args:
        config (ConfigDict): configuration
        location (FilePath): file to write

    Raises:
        ModuleNotFoundError: toml library is required for writing files
    """

    toml = try_import("toml")

    if toml is None:  # pragma: no cover
        message = "Please install the toml library to write TOML files."
        raise ModuleNotFoundError(message)

    with open(location, "w", encoding="utf-8") as _file:
        toml.dump(config, _file)  # type: ignore



"""Typer Configuration Utilities."""

from .callbacks import (
    conf_callback_factory,
    dotenv_conf_callback,
    json_conf_callback,
    toml_conf_callback,
    yaml_conf_callback,
)
from .decorators import use_config, use_json_config, use_toml_config, use_yaml_config
from .loaders import dotenv_loader, ini_loader, json_loader, toml_loader, yaml_loader

__all__ = [
    "dotenv_conf_callback",
    "conf_callback_factory",
    "json_conf_callback",
    "toml_conf_callback",
    "yaml_conf_callback",
    "use_config",
    "use_json_config",
    "use_toml_config",
    "use_yaml_config",
    "dotenv_loader",
    "ini_loader",
    "json_loader",
    "toml_loader",
    "yaml_loader",
]



"""Configuration File Loaders.

These loaders must implement the `typer_config.__typing.ConfigLoader` interface.
"""
from __future__ import annotations

import json
from configparser import ConfigParser
from typing import TYPE_CHECKING, Optional

from .__optional_imports import try_import

if TYPE_CHECKING:  # pragma: no cover
    from .__typing import (
        ConfigDict,
        ConfigDictTransformer,
        ConfigLoader,
        ConfigLoaderConditional,
        TyperParameterValue,
        TyperParameterValueTransformer,
    )


def loader_transformer(
    loader: ConfigLoader,
    loader_conditional: Optional[ConfigLoaderConditional] = None,
    param_transformer: Optional[TyperParameterValueTransformer] = None,
    config_transformer: Optional[ConfigDictTransformer] = None,
) -> ConfigLoader:
    """Configuration loader transformer.

    This allows to transform the input and output of a configuration loader.

    Examples:
        Set a default file to open when none is given:
        ```py
        default_file_loader = loader_transformer(
            yaml_loader,
            param_transformer=lambda param: param if param else "config.yml",
        )
        ```

        Use a subsection of a file:
        ```py
        subsection_loader = loader_transformer(
            yaml_loader,
            config_transformer = lambda config: config["subsection"],
        )
        ```

        Use both transformers to use the `[tool.my_tool]` section from `pyproject.toml`
        by default:
        ```py
        pyproject_loader = loader_transformer(
            toml_loader,
            param_transformer = lambda param: param if param else "pyproject.toml"
            config_transformer = lambda config: config["tool"]["my_tool"],
        )
        ```

    Args:
        loader (ConfigLoader): Loader to transform.
        loader_conditional (Optional[ConfigLoaderConditional], optional): Function
            to determine whether to execute loader. Defaults to None (no-op).
        param_transformer (Optional[TyperParameterValueTransformer], optional): Typer
            parameter transformer. Defaults to None (no-op).
        config_transformer (Optional[ConfigDictTransformer], optional): Config
            dictionary transformer. Defaults to None (no-op).

    Returns:
        ConfigLoader: Transformed config loader.
    """

    def _loader(param_value: TyperParameterValue) -> ConfigDict:
        # Transform input
        if param_transformer is not None:
            param_value = param_transformer(param_value)

        # Decide whether to execute loader
        # NOTE: bad things can happen when `param_value=''`
        # such as `--help` not working
        conf: ConfigDict = {}
        if loader_conditional is None or loader_conditional(param_value):
            conf = loader(param_value)

        # Transform output
        if config_transformer is not None:
            conf = config_transformer(conf)

        return conf

    return _loader


def yaml_loader(param_value: TyperParameterValue) -> ConfigDict:
    """YAML file loader.

    Args:
        param_value (TyperParameterValue): path of YAML file

    Raises:
        ModuleNotFoundError: pyyaml library is not installed

    Returns:
        ConfigDict: dictionary loaded from file
    """

    yaml = try_import("yaml")

    if yaml is None:  # pragma: no cover
        message = "Please install the pyyaml library."
        raise ModuleNotFoundError(message)

    with open(param_value, "r", encoding="utf-8") as _file:
        conf: ConfigDict = yaml.safe_load(_file)

    return conf


def json_loader(param_value: TyperParameterValue) -> ConfigDict:
    """JSON file loader.

    Args:
        param_value (TyperParameterValue): path of JSON file

    Returns:
        ConfigDict: dictionary loaded from file
    """

    with open(param_value, "r", encoding="utf-8") as _file:
        conf: ConfigDict = json.load(_file)

    return conf


def toml_loader(param_value: TyperParameterValue) -> ConfigDict:
    """TOML file loader.

    Args:
        param_value (TyperParameterValue): path of TOML file

    Raises:
        ModuleNotFoundError: toml library is not installed

    Returns:
        ConfigDict: dictionary loaded from file
    """

    # try `tomllib` first
    tomllib = try_import("tomllib")

    if tomllib is not None:
        with open(param_value, "rb") as _file:
            return tomllib.load(_file)

    # couldn't find `tommllib`, so try `toml`
    toml = try_import("toml")

    if toml is None:  # pragma: no cover
        message = "Please install the toml library."
        raise ModuleNotFoundError(message)

    with open(param_value, "r", encoding="utf-8") as _file:
        return toml.load(_file)


def dotenv_loader(param_value: TyperParameterValue) -> ConfigDict:
    """Dotenv file loader.

    Args:
        param_value (TyperParameterValue): path of Dotenv file

    Raises:
        ModuleNotFoundError: python-dotenv library is not installed

    Returns:
        ConfigDict: dictionary loaded from file
    """

    dotenv = try_import("dotenv")

    if dotenv is None:  # pragma: no cover
        message = "Please install the python-dotenv library."
        raise ModuleNotFoundError(message)

    with open(param_value, "r", encoding="utf-8") as _file:
        # NOTE: I'm using a stream here so that the loader
        # will raise an exception when the file doesn't exist.
        conf: ConfigDict = dotenv.dotenv_values(stream=_file)

    return conf


def ini_loader(param_value: TyperParameterValue) -> ConfigDict:
    """INI file loader.

    Note:
        INI files must have sections at the top level.
        You probably want to combine this with `loader_transformer`
        to extract the correct section.
        For example:
        ```py
        ini_section_loader = loader_transformer(
            ini_loader,
            config_transformer=lambda config: config["section"],
        )
        ```

    Args:
        param_value (TyperParameterValue): path of INI file

    Returns:
        ConfigDict: dictionary loaded from file
    """

    ini_parser = ConfigParser()
    with open(param_value, "r", encoding="utf-8") as _file:
        ini_parser.read_file(_file)

    conf: ConfigDict = {
        sect: dict(ini_parser.items(sect)) for sect in ini_parser.sections()
    }

    return conf



"""Data and Function types."""

import sys
from pathlib import Path
from typing import Any, Callable, Dict, Union

from typer import CallbackParam, Context

# Handle some imports based on python version
if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import TypeAlias
else:  # pragma: no cover
    from typing import TypeAlias

# Data types
TyperParameterName: TypeAlias = str
"""Typer CLI parameter name."""

TyperParameterValue: TypeAlias = Any
"""Typer CLI parameter value."""

ConfigDict: TypeAlias = Dict[TyperParameterName, Any]
"""Configuration dictionary to be applied to the click context default map."""


FilePath: TypeAlias = Union[Path, str]
"""File path"""

# Function types
TyperParameterValueTransformer: TypeAlias = Callable[
    [TyperParameterValue], TyperParameterValue
]
"""Typer parameter value transforming function."""

ConfigDictTransformer: TypeAlias = Callable[[ConfigDict], ConfigDict]
"""ConfigDict transforming function."""

ConfigLoader: TypeAlias = Callable[[TyperParameterValue], ConfigDict]
"""Configuration loader function."""

ConfigLoaderConditional: TypeAlias = Callable[[TyperParameterValue], bool]
"""Configuration loader conditional function."""

ConfigParameterCallback: TypeAlias = Callable[
    [Context, CallbackParam, TyperParameterValue], TyperParameterValue
]
"""Typer config parameter callback function."""

ConfigDumper: TypeAlias = Callable[[ConfigDict, FilePath], None]
"""Configuration dumper function."""

TyperCommand: TypeAlias = Callable[..., Any]
"""A function that will be decorated with `typer.Typer().command()`."""

TyperCommandDecorator: TypeAlias = Callable[[TyperCommand], TyperCommand]
"""A decorator applied to a typer command."""



"""Typer Configuration Parameter Callbacks."""

from typer import BadParameter, CallbackParam, Context

from .__typing import ConfigLoader, ConfigParameterCallback, TyperParameterValue
from .loaders import (
    dotenv_loader,
    json_loader,
    loader_transformer,
    toml_loader,
    yaml_loader,
)


def conf_callback_factory(loader: ConfigLoader) -> ConfigParameterCallback:
    """Typer configuration callback factory.

    Args:
        loader (ConfigLoader): Config loader function that takes the value
            passed to the typer CLI and returns a dictionary that is
            applied to the click context's default map.

    Returns:
        ConfigParameterCallback: Configuration parameter callback function.
    """

    def _callback(
        ctx: Context, param: CallbackParam, param_value: TyperParameterValue
    ) -> TyperParameterValue:
        """Generated typer config parameter callback.

        Args:
            ctx (typer.Context): typer context (automatically passed)
            param (typer.CallbackParam): typer callback parameter (automatically passed)
            param_value (TyperParameterValue): parameter value passed to typer
                (automatically passed)

        Raises:
            BadParameter: bad parameter value

        Returns:
            TyperParameterValue: must return back the given parameter
        """
        try:
            conf = loader(param_value)  # Load config file
            ctx.default_map = ctx.default_map or {}  # Initialize the default map
            ctx.default_map.update(conf)  # Merge the config Dict into default_map
        except Exception as ex:  # noqa: BLE001 (reraising to typer framework)
            raise BadParameter(str(ex), ctx=ctx, param=param) from ex
        return param_value

    return _callback


yaml_conf_callback: ConfigParameterCallback = conf_callback_factory(
    loader_transformer(yaml_loader, loader_conditional=lambda param_value: param_value)
)
"""YAML typer config parameter callback.

Args:
    ctx (typer.Context): typer context (automatically passed)
    param (typer.CallbackParam): typer callback parameter (automatically passed)
    param_value (TyperParameterValue): parameter value passed to typer (automatically
        passed)

Raises:
    BadParameter: bad parameter value

Returns:
    TyperParameterValue: must return back the given parameter
"""

json_conf_callback: ConfigParameterCallback = conf_callback_factory(
    loader_transformer(json_loader, loader_conditional=lambda param_value: param_value)
)
"""JSON typer config parameter callback.

Args:
    ctx (typer.Context): typer context (automatically passed)
    param (typer.CallbackParam): typer callback parameter (automatically passed)
    param_value (TyperParameterValue): parameter value passed to typer (automatically
        passed)

Raises:
    BadParameter: bad parameter value

Returns:
    TyperParameterValue: must return back the given parameter
"""


toml_conf_callback: ConfigParameterCallback = conf_callback_factory(
    loader_transformer(toml_loader, loader_conditional=lambda param_value: param_value)
)
"""TOML typer config parameter callback.

Args:
    ctx (typer.Context): typer context (automatically passed)
    param (typer.CallbackParam): typer callback parameter (automatically passed)
    param_value (TyperParameterValue): parameter value passed to typer (automatically
        passed)

Raises:
    BadParameter: bad parameter value

Returns:
    TyperParameterValue: must return back the given parameter
"""

dotenv_conf_callback: ConfigParameterCallback = conf_callback_factory(
    loader_transformer(
        dotenv_loader, loader_conditional=lambda param_value: param_value
    )
)
"""Dotenv typer config parameter callback.

Args:
    ctx (typer.Context): typer context (automatically passed)
    param (typer.CallbackParam): typer callback parameter (automatically passed)
    param_value (TyperParameterValue): parameter value passed to typer (automatically
        passed)

Raises:
    BadParameter: bad parameter value

Returns:
    TyperParameterValue: must return back the given parameter
"""



"""Handle optional imports."""

from functools import lru_cache
from importlib import import_module
from importlib.util import find_spec


@lru_cache()
def try_import(module_name: str):  # noqa: ANN202 (no type for modules)
    """Try to import a module by name.

    Note: caches the imported modules in a `functools.lru_cache`

    Args:
        module_name (str): name of module to import

    Returns:
        Module: imported module
    """
    if find_spec(module_name):
        return import_module(module_name)
    return None



"""Typer Config decorators."""

from __future__ import annotations

from enum import Enum
from functools import wraps
from inspect import Parameter, signature
from typing import TYPE_CHECKING, List, Optional

from typer import Option

from .callbacks import (
    conf_callback_factory,
    dotenv_conf_callback,
    json_conf_callback,
    toml_conf_callback,
    yaml_conf_callback,
)
from .dumpers import json_dumper, toml_dumper, yaml_dumper
from .loaders import (
    dotenv_loader,
    ini_loader,
    json_loader,
    loader_transformer,
    toml_loader,
    yaml_loader,
)

if TYPE_CHECKING:  # pragma: no cover
    from .__typing import (
        ConfigDict,
        ConfigDumper,
        ConfigParameterCallback,
        FilePath,
        TyperCommand,
        TyperCommandDecorator,
        TyperParameterName,
        TyperParameterValue,
    )


def use_config(
    callback: ConfigParameterCallback,
    param_name: TyperParameterName = "config",
    param_help: str = "Configuration file.",
) -> TyperCommandDecorator:
    """Decorator for using configuration on a typer command.

    Usage:
        ```py
        import typer
        from typer_config.decorators import use_config
        from typer_config import yaml_conf_callback # whichever callback to use

        app = typer.Typer()

        @app.command()
        @use_config(yaml_conf_callback)
        def main(...):
            ...
        ```

    Args:
        callback (ConfigParameterCallback): config parameter callback to load
        param_name (TyperParameterName, optional): name of config parameter.
            Defaults to "config".
        param_help (str, optional): config parameter help string.
            Defaults to "Configuration file.".

    Returns:
        TyperCommandDecorator: decorator to apply to command
    """

    def decorator(cmd: TyperCommand) -> TyperCommand:
        # NOTE: modifying a function's __signature__ is dangerous
        # in the sense that it only affects inspect.signature().
        # It does not affect the actual function implementation.
        # So, a caller can be confused how to pass parameters to
        # the function with modified signature.
        sig = signature(cmd)

        config_param = Parameter(
            param_name,
            kind=Parameter.KEYWORD_ONLY,
            annotation=str,
            default=Option("", callback=callback, is_eager=True, help=param_help),
        )

        new_sig = sig.replace(parameters=[*sig.parameters.values(), config_param])

        @wraps(cmd)
        def wrapped(*args, **kwargs):  # noqa: ANN202,ANN002,ANN003
            # NOTE: need to delete the config parameter
            # to match the wrapped command's signature.
            if param_name in kwargs:
                del kwargs[param_name]

            return cmd(*args, **kwargs)

        wrapped.__signature__ = new_sig  # type: ignore

        return wrapped

    return decorator


# default decorators
def use_json_config(
    param_name: TyperParameterName = "config",
    param_help: str = "Configuration file.",
    default_value: Optional[TyperParameterValue] = None,
) -> TyperCommandDecorator:
    """Decorator for using JSON configuration on a typer command.

    Usage:
        ```py
        import typer
        from typer_config.decorators import use_json_config

        app = typer.Typer()

        @app.command()
        @use_json_config()
        def main(...):
            ...
        ```

    Args:
        param_name (TyperParameterName, optional): name of config parameter.
            Defaults to "config".
        param_help (str, optional): config parameter help string.
            Defaults to "Configuration file.".
        default_value (TyperParameterValue, optional): default config parameter value.
            Defaults to None.

    Returns:
        TyperCommandDecorator: decorator to apply to command
    """

    if default_value is not None:
        callback = conf_callback_factory(
            loader_transformer(
                json_loader,
                loader_conditional=lambda param_value: param_value,
                param_transformer=lambda param_value: param_value
                if param_value
                else default_value,
            )
        )
    else:
        callback = json_conf_callback

    return use_config(callback=callback, param_name=param_name, param_help=param_help)


def use_yaml_config(
    param_name: TyperParameterName = "config",
    param_help: str = "Configuration file.",
    default_value: Optional[TyperParameterValue] = None,
) -> TyperCommandDecorator:
    """Decorator for using YAML configuration on a typer command.

    Usage:
        ```py
        import typer
        from typer_config.decorators import use_yaml_config

        app = typer.Typer()

        @app.command()
        @use_yaml_config()
        def main(...):
            ...
        ```

    Args:
        param_name (str, optional): name of config parameter. Defaults to "config".
        param_help (str, optional): config parameter help string.
            Defaults to "Configuration file.".
        default_value (TyperParameterValue, optional): default config parameter value.
            Defaults to None.

    Returns:
        TyperCommandDecorator: decorator to apply to command
    """

    if default_value is not None:
        callback = conf_callback_factory(
            loader_transformer(
                yaml_loader,
                loader_conditional=lambda param_value: param_value,
                param_transformer=lambda param_value: param_value
                if param_value
                else default_value,
            )
        )
    else:
        callback = yaml_conf_callback

    return use_config(callback=callback, param_name=param_name, param_help=param_help)


def use_toml_config(
    param_name: TyperParameterName = "config",
    param_help: str = "Configuration file.",
    default_value: Optional[TyperParameterValue] = None,
) -> TyperCommandDecorator:
    """Decorator for using TOML configuration on a typer command.

    Usage:
        ```py
        import typer
        from typer_config.decorators import use_toml_config

        app = typer.Typer()

        @app.command()
        @use_toml_config()
        def main(...):
            ...
        ```

    Args:
        param_name (str, optional): name of config parameter. Defaults to "config".
        param_help (str, optional): config parameter help string.
            Defaults to "Configuration file.".
        default_value (TyperParameterValue, optional): default config parameter value.
            Defaults to None.

    Returns:
        TyperCommandDecorator: decorator to apply to command
    """

    if default_value is not None:
        callback = conf_callback_factory(
            loader_transformer(
                toml_loader,
                loader_conditional=lambda param_value: param_value,
                param_transformer=lambda param_value: param_value
                if param_value
                else default_value,
            )
        )
    else:
        callback = toml_conf_callback

    return use_config(callback=callback, param_name=param_name, param_help=param_help)


def use_dotenv_config(
    param_name: TyperParameterName = "config",
    param_help: str = "Configuration file.",
    default_value: Optional[TyperParameterValue] = None,
) -> TyperCommandDecorator:
    """Decorator for using dotenv configuration on a typer command.

    Usage:
        ```py
        import typer
        from typer_config.decorators import use_dotenv_config

        app = typer.Typer()

        @app.command()
        @use_dotenv_config()
        def main(...):
            ...
        ```

    Args:
        param_name (str, optional): name of config parameter. Defaults to "config".
        param_help (str, optional): config parameter help string.
            Defaults to "Configuration file.".
        default_value (TyperParameterValue, optional): default config parameter value.
            Defaults to None.

    Returns:
        TyperCommandDecorator: decorator to apply to command
    """

    if default_value is not None:
        callback = conf_callback_factory(
            loader_transformer(
                dotenv_loader,
                loader_conditional=lambda param_value: param_value,
                param_transformer=lambda param_value: param_value
                if param_value
                else default_value,
            )
        )
    else:
        callback = dotenv_conf_callback

    return use_config(callback=callback, param_name=param_name, param_help=param_help)


def use_ini_config(
    section: List[str],
    param_name: TyperParameterName = "config",
    param_help: str = "Configuration file.",
    default_value: Optional[TyperParameterValue] = None,
) -> TyperCommandDecorator:
    """Decorator for using INI configuration on a typer command.

    Usage:
        ```py
        import typer
        from typer_config.decorators import use_ini_config

        app = typer.Typer()

        @app.command()
        @use_ini_config(["section", "subsection"])
        def main(...):
            ...
        ```

    Args:
        section (List[str]): List of nested sections to access in the INI file.
        param_name (str, optional): name of config parameter. Defaults to "config".
        param_help (str, optional): config parameter help string.
            Defaults to "Configuration file.".
        default_value (TyperParameterValue, optional): default config parameter value.
            Defaults to None.

    Returns:
        TyperCommandDecorator: decorator to apply to command
    """

    def _get_section(_section: List[str], config: ConfigDict) -> ConfigDict:
        for sect in _section:
            config = config.get(sect, {})

        return config

    if default_value is not None:
        callback = conf_callback_factory(
            loader_transformer(
                ini_loader,
                loader_conditional=lambda param_value: param_value,
                param_transformer=lambda param_value: param_value
                if param_value
                else default_value,
                config_transformer=lambda config: _get_section(section, config),
            )
        )
    else:
        callback = conf_callback_factory(
            loader_transformer(
                ini_loader,
                loader_conditional=lambda param_value: param_value,
                config_transformer=lambda config: _get_section(section, config),
            )
        )

    return use_config(callback=callback, param_name=param_name, param_help=param_help)


def dump_config(dumper: ConfigDumper, location: FilePath) -> TyperCommandDecorator:
    """Decorator for dumping a config file with parameters
    from an invocation of a typer command.

    Usage:
        ```py
        import typer
        from typer.decorators import dump_config

        app = typer.Typer()

        @app.command()
        # NOTE: @dump_config MUST BE AFTER @app.command()
        @dump_config(yaml_dumper, "config_dump_dir/params.yaml")
        def cmd(...):
            ...
        ```

    Args:
        dumper (ConfigDumper): config file dumper
        location (FilePath): config file to write

    Returns:
        TyperCommandDecorator: command decorator
    """

    def decorator(cmd: TyperCommand) -> TyperCommand:
        @wraps(cmd)
        def inner(*args, **kwargs):  # noqa: ANN202,ANN002,ANN003
            # get a dictionary of the passed args
            bound_args = signature(cmd).bind(*args, **kwargs).arguments

            # convert enums to their values
            # NOTE: bound_args shouldn't be nested in the typer
            # framework, so top level iteration should be fine.
            for key, val in bound_args.items():
                if isinstance(val, Enum):
                    bound_args[key] = val.value

            # dump passed args
            dumper(bound_args, location)

            # run original command
            return cmd(*args, **kwargs)

        return inner

    return decorator


def dump_json_config(location: FilePath) -> TyperCommandDecorator:
    """Decorator for dumping a JSON file with parameters
    from an invocation of a typer command.

    Usage:
        ```py
        import typer
        from typer.decorators import dump_json_config

        app = typer.Typer()

        @app.command()
        # NOTE: @dump_json_config MUST BE AFTER @app.command()
        @dump_json_config("config_dump_dir/params.json")
        def cmd(...):
            ...
        ```

    Args:
        location (FilePath): config file to write

    Returns:
        TyperCommandDecorator: command decorator
    """
    return dump_config(dumper=json_dumper, location=location)


def dump_yaml_config(location: FilePath) -> TyperCommandDecorator:
    """Decorator for dumping a YAML file with parameters
    from an invocation of a typer command.

    Usage:
        ```py
        import typer
        from typer.decorators import dump_yaml_config

        app = typer.Typer()

        @app.command()
        # NOTE: @dump_yaml_config MUST BE AFTER @app.command()
        @dump_yaml_config("config_dump_dir/params.yml")
        def cmd(...):
            ...
        ```

    Args:
        location (FilePath): config file to write

    Returns:
        TyperCommandDecorator: command decorator
    """
    return dump_config(dumper=yaml_dumper, location=location)


def dump_toml_config(location: FilePath) -> TyperCommandDecorator:
    """Decorator for dumping a TOML file with parameters
    from an invocation of a typer command.

    Usage:
        ```py
        import typer
        from typer.decorators import dump_toml_config

        app = typer.Typer()

        @app.command()
        # NOTE: @dump_toml_config MUST BE AFTER @app.command()
        @dump_toml_config("config_dump_dir/params.toml")
        def cmd(...):
            ...
        ```

    Args:
        location (FilePath): config file to write

    Returns:
        TyperCommandDecorator: command decorator
    """
    return dump_config(dumper=toml_dumper, location=location)
